# -*- coding: utf-8 -*-

"""
HTML timetable events viewer.
"""

__all__ = ("EventsViewer",)

from itertools import chain
from django.utils.html import escape
from django.utils.safestring import mark_safe

from datetime import timedelta


GRANULARITY = 30  # In minutes
DEFAULT_START = 8 * (60 // GRANULARITY)  # 8:00
DEFAULT_END = 18 * (60 // GRANULARITY)  # 18:00
DAYS = ("Lundi", "Mardi", "Mercredi", "Jeudi", "Vendredi", "Samedi", "Dimanche")


def _conv_time(value):
    if value is None:
        return None

    return (value.hour * 60 + value.minute) // GRANULARITY


class HTMLTableBuilder(object):
    """
    A helper class to build the rows and cells of the table.
    """

    def __init__(self, output):
        self._out = output
        self._stack = []

    def tr(self, **kwargs):
        if self._stack:
            raise Exception("Already in row!")

        self.fix_dict(kwargs)

        self._out.append("<tr")
        for key, val in kwargs.items():
            self._out.append(f' {key}="{escape(val)}"')
        self._out.append(">\n")

        self._stack.append("\n</tr>\n")

        return self

    def tag(self, name, **kwargs):
        if not self._stack:
            raise Exception("Not in row!")

        self.fix_dict(kwargs)

        self._out.append(f"<{name}")
        for key, val in kwargs.items():
            self._out.append(f' {key}="{escape(val)}"')
        self._out.append(">")

        self._stack.append(f"</{name}>")

        return self

    def single(self, name, **kwargs):
        if not self._stack:
            raise Exception("Not in row!")

        self.fix_dict(kwargs)

        self._out.append(f"<{name}")
        for key, val in kwargs.items():
            self._out.append(f' {key}="{escape(val)}"')
        self._out.append(" />")
        self._stack.append("")

        return self

    @staticmethod
    def fix_dict(items_dict):
        if "rowspan" in items_dict and items_dict["rowspan"] == 1:
            del items_dict["rowspan"]

        if "klass" not in items_dict:
            return

        klass = items_dict.pop("klass")
        if isinstance(klass, list):
            if not klass:
                return

            klass = " ".join(klass)

        items_dict["class"] = klass

    def text(self, value):
        self._out.append(escape(value))

    def __getattr__(self, name):
        def wrapper(**kwargs):
            if name in ("br",):
                return self.single(name, **kwargs)
            else:
                return self.tag(name, **kwargs)

        return wrapper

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._out.append(self._stack.pop())

        return False


class EventsCellMatrix(object):
    """
    Matrix of event cells.
    Mainly used to properly iterate over them
    """

    __slots__ = ("_items", "_days")

    def __init__(self, day_count, start_pos, end_pos):
        self._items = items = []
        self._days = day_count

        for pos in range(start_pos, end_pos):
            for _ in range(day_count):
                items.append(EventsNode(pos))

    def get_height(self):
        return len(self._items) // self._days

    def get_row(self, row):
        return self._items[row * self._days : (row + 1) * self._days]

    def get_day(self, day):
        return self._items[day :: self._days]

    def __iter__(self):
        for row in range(self.get_height()):
            yield self.get_row(row)


def cmp(a, b):
    return (a > b) - (a < b)  # équivalent cmp en python 3


class EventsNode(object):
    """
    This represents a "cell" of the timetable.
    The public functions defined here are meant to be called on the upper left node of
    the table.
    """

    __slots__ = ("_pos", "_events", "_span", "_first", "_last", "_limitless")

    def __init__(self, pos):
        self._pos = pos
        self._events = []
        self._span = 1
        self._first = False
        self._last = False
        self._limitless = False

    def set_span(self, span, first, last):
        """
        Sets the row span of the cell.
        Also indicates to the cell if it will be the first displayed cell in the column,
        and/or the last one.
        """

        self._span = span
        self._first = first
        self._last = last

    def add_event(self, event, limitless):
        """
        Adds the event to this cell.
        limitless: True if the event starts before or end after this day.
        """

        self._limitless = self._limitless or limitless

        self._events.append(event)

    def output_to(self, builder):
        """
        Outputs the HTML representation of this node in the HTML builder
        """

        if self._span == 0:
            return

        if self._events:
            klass = ["event"]

            categories = set(event.category for event in self._events)
            if len(categories) == 1:
                klass.append(categories.pop())

            if self._limitless:
                if self._first:
                    klass.append("continue_first")

                if self._last:
                    klass.append("continue_last")

        else:
            klass = []

        with builder.td(klass=klass, rowspan=self._span):
            extra_data = []

            for i, event in enumerate(self._events):
                if i > 0:
                    with builder.br():
                        pass
                    with builder.em():
                        builder.text("ou")
                    with builder.br():
                        pass

                with builder.strong():
                    builder.text(event.title)

                subtext = []

                if event.type:
                    subtext.append(event.type)

                if event.locations:
                    if len(event.locations) < 5:
                        subtext.append(", ".join(event.locations))
                    else:
                        extra_data.append("Salles : " + ", ".join(event.locations))

                if event.people:
                    extra_data.append("Avec " + ", ".join(event.people))

                if subtext:
                    with builder.br():
                        pass
                    builder.text(" — ".join(subtext))

            if extra_data:
                with builder.aside(klass="extra"):
                    for i, text in enumerate(extra_data):
                        if i > 0:
                            with builder.br():
                                pass
                        builder.text(text)

    def __len__(self):
        return len(self._events)

    def __cmp__(self, other):
        return cmp(self._events, other._events)


class EventsViewer(object):
    def __init__(self, events_lists, first_day):
        self._matrix = None
        self._day_count = 0
        self._start = 0
        self._start_day = first_day

        first_start = DEFAULT_START
        last_end = DEFAULT_END

        day_events = [[] for _ in range(7)]

        for event in chain(*events_lists):
            start = _conv_time(event.start)
            end = _conv_time(event.end)
            day_events[event.date.weekday()].append(event)

            # Ensure that limitless events have a bit of room to be displayed
            if start is None and end is not None:
                start = max(0, end - 2)
            elif start is not None and end is None:
                end = min(1440 // GRANULARITY, start + 2)

            if start is not None and start < first_start:
                first_start = start

            if end is not None and end > last_end:
                last_end = end

        if day_events[6]:
            day_count = 7
        elif day_events[5]:
            day_count = 6
        else:
            day_count = 5

        self._matrix = EventsCellMatrix(day_count, first_start, last_end)
        self._day_count = day_count
        self._start = first_start

        for day, event_list in enumerate(day_events[:day_count]):
            day_cells = self._matrix.get_day(day)

            for event in event_list:
                limitless = False

                if event.start is None:
                    limitless = True
                    start = 0
                else:
                    start = _conv_time(event.start) - first_start

                if event.end is None:
                    limitless = True
                    end = last_end - first_start
                else:
                    end = _conv_time(event.end) - first_start

                for cell in day_cells[start:end]:
                    cell.add_event(event.data, limitless)

        # Remove the duplicate cells, adjust spans
        count = self._matrix.get_height()
        for day in range(day_count):
            cells = self._matrix.get_day(day)

            ref_index = 0

            while ref_index < count:
                ref_is_first = ref_index == 0
                ref_cell = cells[ref_index]

                if not ref_cell:  # No events in this cell, not merged with another
                    ref_index += 1
                    continue

                cur_index = ref_index + 1

                while cur_index < count:
                    cur_cell = cells[cur_index]
                    if cur_cell == ref_cell:
                        cur_cell.set_span(0, False, False)
                        cur_index += 1
                    else:
                        break

                ref_is_last = cur_index == count

                # cur_index - ref_index == 1 if the next cell is different from the ref
                ref_cell.set_span(cur_index - ref_index, ref_is_first, ref_is_last)
                ref_index = cur_index

    def __str__(self):
        result = []

        builder = HTMLTableBuilder(result)
        with builder.tr(klass="days"):
            with builder.th(klass="hour"):
                pass

            for i, week_day in enumerate(DAYS[: self._day_count]):
                with builder.th():
                    month_day = self._start_day + timedelta(days=i)
                    builder.text(f"{week_day} {month_day.day}")

        for i, cell_row in enumerate(chain(self._matrix, [[]])):
            with builder.tr():
                with builder.th(klass="hour"):
                    with builder.span():
                        minutes = (self._start + i) * GRANULARITY
                        hours = minutes // 60
                        minutes = minutes % 60
                        builder.text(f"{hours}:{minutes:02d}")

                for cell in cell_row:
                    cell.output_to(builder)

        return mark_safe("".join(result))
